import PIL
import io

from matplotlib.pyplot import savefig
from matplotlib.pyplot import show

from .figures import Figures
from .. import defaults

class Animate(Figures):
    
    def __init__(self, data_objects, **kwargs):
        Figures.__init__(self, data_objects, **kwargs)
        defaults.kwargs(self, kwargs)
    
    def create_animations(self, **kwargs):
        self.process_data_objects()
        self.set_figure_objects(**kwargs)
        self.prepare_animation_settings()
        self.animate_data_objects()

    def prepare_animation_settings(self):
        self.output = None
        self.set_animation_axis_limits()
        self.set_figure_animation_settings()

    def set_animation_axis_limits(self):
        for figure_obj in self.figure_objects:
            figure_obj.set_animation_axis_limits()

    def set_figure_animation_settings(self):
        for figure_obj in self.figure_objects:
            figure_obj.maximise = self.maximise
            figure_obj.figure_size = self.figure_size

    def animate_data_objects(self):
        for figure_obj in self.figure_objects:
            self.animate_figure(figure_obj)

    def animate_figure(self, figure_obj):
        frame_count = figure_obj.get_frame_count()
        self.set_all_data_values(figure_obj)
        frames = self.get_frames(figure_obj, frame_count)
        self.save_animation(frames)

    def set_all_data_values(self, figure_obj):
        figure_obj.all_data_values = [data_obj.get_data_values()
                                      for data_obj in figure_obj.data_objects]

    def get_frames(self, figure_obj, frame_count):
        frames = [self.get_frame(figure_obj, index)
                  for index in range(frame_count)]
        return frames

    def get_frame(self, figure_obj, index):
        buffer = self.get_buffer(figure_obj, index)
        image = self.get_image(buffer, figure_obj)
        return image

    def get_buffer(self, figure_obj, index):
        figure_obj.set_data_value(index)
        figure_obj.create_figure()
        buffer = io.BytesIO()
        return buffer

    def get_image(self, buffer, figure_obj):
        savefig(buffer, dpi=figure_obj.fig.dpi)
        buffer.seek(0)
        image = PIL.Image.open(buffer)
        image = image.resize(figure_obj.figure_size_pixels)
        return image

    def save_animation(self, frames):
        animation_path = f"{self.animation_name}.{self.format}"
        frames[0].save(animation_path, loop=self.loop, save_all=True,
                       append_images=frames[1:], duration=self.duration)

defaults.load(Animate)



import os

from matplotlib.pyplot import savefig

def save_figure(plot_obj):
    path = get_figure_path(plot_obj)
    plots_obj = plot_obj.plots_obj
    save_fig(path, plots_obj)
    plt.close()

def get_figure_path(plot_obj):
    file_name = get_file_name(plot_obj)
    file_name = f"{file_name}.{plot_obj.plots_obj.format}"
    path = os.path.join(plot_obj.plots_obj.path, file_name)
    return path

def get_file_name(plot_obj):
    if len(plot_obj.plots_obj.lines_object_groups) == 1:
        return get_base_file_name(plot_obj.plots_obj)
    else:
        return get_numbered_file_name(plot_obj)

def get_base_file_name(plot_objs):
    if plot_objs.title is None:
        return "Figure"
    else:
        return str(plot_objs.title)

def get_numbered_file_name(plot_obj):
    file_name = get_base_file_name(plot_obj.plot_objs)
    file_name = f"{file_name} {plot_obj.plot_index + 1}"
    return file_name

def save_fig(path, plots_obj):
    savefig(path,
            dpi=plots_obj.dpi,
            format=plots_obj.format,
            metadata=plots_obj.metadata,
            bbox_inches=plots_obj.bbox_inches,
            pad_inches=plots_obj.pad_inches,
            facecolor=plots_obj.facecolor,
            edgecolor=plots_obj.edgecolor,
            backend=plots_obj.backend)



import json

from ..utils.paths import load_json
from ..utils.paths import make_folder_path

class ProcessKwargs():

    def __init__(self, obj, *args, **kwargs):
        self.set_obj(obj)
        self.set_kwargs(args, kwargs)
        self.try_process_kwargs()

    def set_obj(self, obj):
        if isinstance(obj, dict):
            raise Exception("You need to pass the object instance into 'kwargs' function")
        else:
            self.obj = obj

    def set_kwargs(self, args, kwargs):
        if len(args) > 0 and isinstance(args[0], dict):
            self.kwargs = args[0]
        else:
            self.kwargs = kwargs

    def try_process_kwargs(self):
        if hasattr(self.obj, "defaults_path"):
            self.process_kwargs()

    def process_kwargs(self):
        self.key_words_to_add = []
        self.process_key_words()
        self.add_key_words_to_defaults()

    def process_key_words(self):
        for key_word, value in self.kwargs.items():
            self.process_key_word(key_word, value)

    def process_key_word(self, key_word, value):
        setattr(self.obj, key_word, value)
        if key_word not in self.obj.defaults:
            self.key_words_to_add.append(key_word)

    def add_key_words_to_defaults(self):
        if self.add_extra_defaults_to_file():
            self.do_add_key_words_to_defaults()

    def add_extra_defaults_to_file(self):
        if "add_input_kwargs_to_file" in self.obj.defaults:
            return self.obj.defaults["add_input_kwargs_to_file"]
        else:
            return False

    def do_add_key_words_to_defaults(self):
        original_file_contents = load_json(self.obj.defaults_path, ignore_empty_or_none=True)
        new_file_contents = self.add_key_words_to_file_contents(original_file_contents)
        self.save_file_contents(original_file_contents, new_file_contents)

    def add_key_words_to_file_contents(self, original_file_contents):
        new_file_contents = dict(original_file_contents)
        for key_word in self.key_words_to_add:
            if key_word not in original_file_contents:
                new_file_contents.update({key_word: None})
        return new_file_contents

    def save_file_contents(self, original_file_contents, new_file_contents):
        if len(original_file_contents) != len(new_file_contents):
            make_folder_path(self.obj.defaults_path)
            with open(self.obj.defaults_path, "w") as file:
                json.dump(new_file_contents, file, indent=2)



import numpy as np
from matplotlib.colors import hsv_to_rgb

from ... import defaults
from .data import Data
from .line import Line

class Lines(Data):

    def __init__(self, line_objects, **kwargs):
        Data.__init__(self, **kwargs)
        defaults.kwargs(self, kwargs)
        self.set_line_objects(line_objects)
        self.count = len(self.line_objects)

    def set_line_objects(self, line_objects):
        if isinstance(line_objects, Line):
            self.line_objects = [line_objects]
        else:
            self.set_line_objects_multiple(line_objects)

    def set_line_objects_multiple(self, line_objects):
        if np.all([isinstance(line_obj, Line)
                   for line_obj in line_objects]):
            self.line_objects = list(line_objects)
        else:
            self.bad_data_objects_exception()

    def bad_data_objects_exception(self):
        message = ("When creating a lines object you must pass "
                   "in an instance of Line or an iterable of "
                   "instances of Line")
        raise TypeError(message)

    def set_rainbow_lines(self, saturation=1, value=1):
        self.set_colours(saturation, value)
        for line_obj, colour in zip(self.line_objects, self.colours):
            line_obj.colour = colour

    def set_colours(self, saturation, value):
        hues = np.linspace(0, 1, self.count + 1)[:self.count]
        saturations = np.ones(self.count)*saturation
        values = np.ones(self.count)*value
        hsv_tuples = np.array(list(zip(hues, saturations, values)))
        self.colours = hsv_to_rgb(hsv_tuples)

defaults.load(Lines)



import traceback
import inspect
import os
import json

from ..utils.paths import make_file
from ..utils.paths import load_json

class LoadDefaults():

    def __init__(self, cls):
        self.cls = cls
        self.set_defaults_path()
        self.set_defaults()

    def set_defaults_path(self):
        module_path = traceback.extract_stack()[-3].filename
        parent_path, module_file_name = os.path.split(module_path)
        defaults_file_name = f"{self.cls.__name__}.txt"
        self.cls.defaults_path = os.path.join(parent_path, "Default Settings", defaults_file_name)

    def set_defaults(self):
        self.cls.defaults = {}
        self.add_defaults_from_ancestors()
        self.add_defaults_from_file()
        self.set_default_values()

    def add_defaults_from_ancestors(self):
        ancestors = self.cls.__mro__[-2:0:-1]
        for ancestor in ancestors:
            self.add_defaults_from_ancestor(ancestor)

    def add_defaults_from_ancestor(self, ancestor):
        for key_word, value in ancestor.defaults.items():
            self.cls.defaults.update({key_word: value})

    def add_defaults_from_file(self):
        defaults_from_file = load_json(self.cls.defaults_path, ignore_empty_or_none=True)
        for key_word, value in defaults_from_file.items():
            self.cls.defaults.update({key_word: value})

    def set_default_values(self):
        for parameter_name, parameter_value in self.cls.defaults.items():
            setattr(self.cls, parameter_name, parameter_value)



from matplotlib.pyplot import get_current_fig_manager

def maximise_figure():
    mng = get_current_fig_manager()
    maximise_figure_attempt_1(mng)

def maximise_figure_attempt_1(mng):
    try:
        mng.resize(*mng.window.maxsize())
    except:
        maximise_figure_attempt_2(mng)

def maximise_figure_attempt_2(mng):
    try:
        mng.window.fullscreen()
    except:
        maximise_figure_attempt_3(mng)

def maximise_figure_attempt_3(mng):
    try:
        mng.window.state('zoomed')
    except:
        maximise_figure_attempt_4(mng)

def maximise_figure_attempt_4(mng):
    try:
        mng.full_screen_toggle()
    except:
        print("Could not maximise figure window")



__author__ = "Henry Ginn"
__date__ = "2023/06/05"
__version__ = "1.0"

from .loaddefaults import LoadDefaults as load
from .processkwargs import ProcessKwargs as kwargs
from .inherit import inherit
from .docs import docs

docs()



import matplotlib.pyplot as plt
import matplotlib.font_manager
import numpy as np

from ... import defaults
from ...utils.dicts import remove_none_values

class Plot():

    @classmethod
    def set_function_dict(cls):
        pass
    
    def __init__(self, figure_obj, ax, data_obj, **kwargs):
        self.set_figure_obj(figure_obj)
        self.ax = ax
        self.data_obj = data_obj
        defaults.kwargs(self, kwargs)
        self.set_kwargs()

    def set_kwargs(self):
        self.inherit_kwargs()
        self.set_font_kwargs()

    def inherit_kwargs(self):
        attributes = ["title_fontname", "title_fontsize",
                      "title_color", "title_verticalalignment",
                      "title_horizontalalignment", "title_y",
                      "title_pad", "title_loc",
                      "axis_fontname", "axis_fontsize",
                      "axis_color", "axis_labelpad", "axis_loc",
                      "x_axis_rotation", "y_axis_rotation", "z_axis_rotation"]
        defaults.inherit(self.data_obj, self.figure_obj, attributes)
        
    def set_font_kwargs(self):
        self.set_title_fontdict()
        self.set_axis_fontdict()

    def set_title_fontdict(self):
        kwargs = {"fontname": self.data_obj.title_fontname,
                  "fontsize": self.data_obj.title_fontsize,
                  "color": self.data_obj.title_color,
                  "verticalalignment": self.data_obj.title_verticalalignment,
                  "horizontalalignment": self.data_obj.title_horizontalalignment}
        self.title_fontdict = remove_none_values(kwargs)

    def set_axis_fontdict(self):
        kwargs = {"fontname": self.data_obj.axis_fontname,
                  "fontsize": self.data_obj.axis_fontsize,
                  "color": self.data_obj.axis_color}
        self.axis_fontdict = remove_none_values(kwargs)

    def set_figure_obj(self, figure_obj):
        self.figure_obj = figure_obj
        self.figures_obj = figure_obj.figures_obj

    def create_plot(self):
        self.plot_data()
        self.set_title()
        self.add_legend()
        self.add_axis_labels()

    def set_title(self):
        if self.data_obj.title is not None:
            self.ax.set_title(self.data_obj.title, **self.title_fontdict,
                              loc=self.data_obj.title_loc, y=self.data_obj.title_y,
                              pad=self.data_obj.title_pad)
    
    def add_legend(self):
        if not self.figures_obj.universal_legend:
            if self.data_obj.legend:
                self.ax.legend(loc=self.data_obj.legend_loc)

    def add_axis_labels(self):
        pass

    def add_x_label(self):
        if self.data_obj.x_label is not None:
            self.ax.set_xlabel(self.data_obj.x_label,
                               **self.axis_fontdict)

    def add_y_label(self):
        if self.data_obj.y_label is not None:
            self.ax.set_ylabel(self.data_obj.y_label,
                               **self.axis_fontdict)

    def add_z_label(self):
        if self.data_obj.z_label is not None:
            self.ax.set_zlabel(self.data_obj.z_label,
                               **self.axis_fontdict)

defaults.load(Plot)



import numpy as np

from ... import defaults
from .data import Data

class Surface(Data):

    def __init__(self, x_values, y_values, z_values, **kwargs):
        Data.__init__(self, **kwargs)
        self.set_xyz_values(x_values, y_values, z_values)
        defaults.kwargs(self, kwargs)

    def set_xyz_values(self, x_values, y_values, z_values):
        self.x_values = x_values
        self.y_values = y_values
        self.z_values = z_values

    def set_animation_axis_limits(self):
        max_z = np.max(self.z_values)
        min_z = np.min(self.z_values)
        self.z_limits = [min_z, max_z]

    def get_frame_count(self):
        if len(self.z_values.shape) == 3:
            return self.z_values.shape[0]
        else:
            raise ValueError("z_values needs 3 dimensions to animate")

    def get_data_values(self):
        return np.copy(self.z_values)

    def set_data_value(self, data_value):
        self.z_values = data_value

defaults.load(Surface)
